#!/usr/bin/env ruby

require 'date'
require "optparse"
require "ostruct"
require "xmlrpc/client"

# A couple things that still need to be done:
#   - Add a quiet mode
#   - Option for the event builders to sprout a canvas
#   - Make disk writing optional
#   - Validate configurations looking for simple problems
#   - Calculate the optimum number of buffers for each process

# It'd be nice if we were using a newer version of ruby (1.9.3) that had
# better string format substitution.  For the time being we'll setup our
# string constants using the newer convention and do the ugly substitutions.
EB_CONFIG = "\
BEGIN_PROLOG
  huff:
  {
    module_type: DS50Compression
    raw_label: daq
    want_bins: false
    perf_print: false
    use_diffs: true
    record_compression: false
  }

  huffdiff_1720: @local::huff
  huffdiff_1720.instance_name: V1720
  huffdiff_1720.table_file: \"table_daqV1720_huff_diff.txt\"
  huffdiff_1724: @local::huff
  huffdiff_1724.instance_name: V1724
  huffdiff_1724.table_file: \"table_daqV1724_huff_diff.txt\"

  unhuff:
  {
    module_type: DS50Decompression
  }

  dunhuff_1720: @local::unhuff
  dunhuff_1720.instance_name: V1720
  dunhuff_1720.compressed_label: huffdiffV1720
  dunhuff_1720.table_file: \"table_daqV1720_huff_diff.txt\"
  dunhuff_1724: @local::unhuff
  dunhuff_1724.instance_name: V1724
  dunhuff_1724.compressed_label: huffdiffV1724
  dunhuff_1724.table_file: \"table_daqV1724_huff_diff.txt\"
END_PROLOG
services: {
  scheduler: {
    fileMode: NOMERGE
  }
  user: {
    NetMonTransportServiceInterface: {
      service_provider: NetMonTransportService
      first_data_receiver_rank: %{ag_rank}
      mpi_buffer_count: 4
      max_fragment_size_words: 2097152
    }
  }
  Timing: { summaryOnly: true }
  #SimpleMemoryCheck: { }
}
daq: {
  max_fragment_size_words: 2097152
  event_builder: {
    mpi_buffer_count: %{buffer_count}
    first_fragment_receiver_rank: 0
    fragment_receiver_count: %{total_frs}
    expected_fragments_per_event: %{total_frs}
    use_art: true
    print_event_store_stats: true
  }
}
outputs: {
  %{netmon_output}netMonOutput: {
  %{netmon_output}  module_type: NetMonOutput
  %{netmon_output}  %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  %{netmon_output}}
  %{root_output}normalOutput: {
  %{root_output}  module_type: RootOutput
  %{root_output}  fileName: \"%{output_file}\"
  %{root_output}  compressionLevel: 0
  %{root_output}  %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  %{root_output}}
  dumper:
  {
    module_type: FileDumperOutput
    wantProductFriendlyClassName: true
  }
}

physics: {
  analyzers: {
    app: {
      module_type: RootApplication
      force_new: true
    }
    wf: {
      module_type: WFViewer
      prescale: 20
      digital_sum_only: false
    }
    baseline: {
      module_type: MonitorBaseline
      random_only: true
    }
    trigger: {
      module_type: MonitorTrigger
    }
    ser: {
      module_type: MonitorSer
      start_us: 0.0
      stop_us: 0.1
      laser_only: true
      #single_channel: 0
      range: [ -1000, 100 ]
    }
  }

  producers: {
    # compress the raw
    huffdiffV1720: @local::huffdiff_1720  
    huffdiffV1724: @local::huffdiff_1724 

    # uncompress the raw
    dunhuffV1720: @local::dunhuff_1720
    dunhuffV1724: @local::dunhuff_1724
  }
  %{enable_1720_compression}p1: [ huffdiffV1720 ]
  %{enable_1724_compression}p1: [ huffdiffV1724 ]
  %{enable_172x_compression}p1: [ huffdiffV1720, huffdiffV1724 ]
  %{enable_compression}trigger_paths: [ p1 ]

  a1: [ app, wf, trigger, ser ] #, baseline, ser ]
  debug: [ dumper ]

  %{netmon_output}my_output_modules: [ netMonOutput ]
  %{root_output}my_output_modules: [ normalOutput ]
  %{output_and_onmon}end_paths: [ a1, my_output_modules ]
  %{output_only}end_paths: [ my_output_modules ]
  %{onmon_only}end_paths: [ a1 ]
}
source: {
  module_type: RawInput
  waiting_time: 20
  resume_after_timeout: true
  fragment_type_map: [[1, \"missed\"], [2, \"V1495\"], [3, \"V1720\"], [4, \"V1724\"], [5, \"V1190\"]]
}
process_name: DAQ"

AG_CONFIG = "\
BEGIN_PROLOG
  huff:
  {
    module_type: DS50Compression
    raw_label: daq
    want_bins: false
    perf_print: false
    use_diffs: true
    record_compression: false
  }

  huffdiff_1720: @local::huff
  huffdiff_1720.instance_name: V1720
  huffdiff_1720.table_file: \"table_daqV1720_huff_diff.txt\"
  huffdiff_1724: @local::huff
  huffdiff_1724.instance_name: V1724
  huffdiff_1724.table_file: \"table_daqV1724_huff_diff.txt\"

  unhuff:
  {
    module_type: DS50Decompression
  }

  dunhuff_1720: @local::unhuff
  dunhuff_1720.instance_name: V1720
  dunhuff_1720.compressed_label: huffdiffV1720
  dunhuff_1720.table_file: \"table_daqV1720_huff_diff.txt\"
  dunhuff_1724: @local::unhuff
  dunhuff_1724.instance_name: V1724
  dunhuff_1724.compressed_label: huffdiffV1724
  dunhuff_1724.table_file: \"table_daqV1724_huff_diff.txt\"
END_PROLOG
services: {
  scheduler: {
    fileMode: NOMERGE
  }
  user: {
    NetMonTransportServiceInterface: {
      service_provider: NetMonTransportService
      max_fragment_size_words: 2097152
    }
  }
  Timing: { summaryOnly: true }
  #SimpleMemoryCheck: { }
}
daq: {
  max_fragment_size_words: 2097152
  aggregator: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    expected_events_per_bunch: %{bunch_size}
    print_event_store_stats: true
  }
}
source: {
  module_type: NetMonInput
}
outputs: {
  normalOutput: {
    module_type: RootOutput
    fileName: \"%{output_file}\"
    compressionLevel: 0
    %{drop_uncompressed}outputCommands: [ \"keep *\", \"drop artdaq::Fragments_daq_V1720_*\", \"drop artdaq::Fragments_daq_V1724_*\" ]
  }
  dumper:
  {
    module_type: FileDumperOutput
    wantProductFriendlyClassName: true
  }
}
physics: {
  analyzers: {
    app: {
      module_type: RootApplication
      force_new: true
    }
    wf: {
      module_type: WFViewer
      prescale: 20
      digital_sum_only: false
    }
    baseline: {
      module_type: MonitorBaseline
      random_only: true
    }
    trigger: {
      module_type: MonitorTrigger
    }
    ser: {
      module_type: MonitorSer
      start_us: 0.0
      stop_us: 0.1
      laser_only: true
      #single_channel: 0
      range: [ -400, 100 ]
    }
  }

  producers: {
    # compress the raw
    huffdiffV1720: @local::huffdiff_1720  
    huffdiffV1724: @local::huffdiff_1724 

    # uncompress the raw
    dunhuffV1720: @local::dunhuff_1720
    dunhuffV1724: @local::dunhuff_1724
  }
  %{enable_1720_compression}p1: [ huffdiffV1720 ]
  %{enable_1724_compression}p1: [ huffdiffV1724 ]
  %{enable_172x_compression}p1: [ huffdiffV1720, huffdiffV1724 ]
  %{enable_compression}trigger_paths: [ p1 ]

  a1: [ app, wf, trigger, ser ] #, baseline, ser ]
  debug: [ dumper ]

  %{root_output}my_output_modules: [ normalOutput ]
  %{output_and_onmon}end_paths: [ a1, my_output_modules ]
  %{output_only}end_paths: [ my_output_modules ]
  %{onmon_only}end_paths: [ a1 ]
}
process_name: DAQAG"

V1495_CONFIG = "\
daq: {
  max_fragment_size_words: 2097152
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1495Driver
    link_type: PCIE
    link_number: %{link_number}
    vme_base: 0x01000000
    pulser_frequency: %{pulser_rate}
    laser_frequency: %{laser_rate}
    random_triggers: %{random_trigger_enable}
    acquisition_gate_us: %{gate_width}
    low_threshold: %{low_threshold}
    %{fnal_mode}fnal_mode: true
    generator_ds50: {
      fragment_id: %{last_fragment_id}
    }
    rt_priority: 2
  }
}"

V1720_CONFIG = "\
BEGIN_PROLOG
def_offset: 0x2700
def_channels: [ { channel: 0 enable: true dc_offset: @local::def_offset },
                { channel: 1 enable: true dc_offset: @local::def_offset },
                { channel: 2 enable: true dc_offset: @local::def_offset },
                { channel: 3 enable: true dc_offset: @local::def_offset },
                { channel: 4 enable: true dc_offset: @local::def_offset },
                { channel: 5 enable: true dc_offset: @local::def_offset },
                { channel: 6 enable: true dc_offset: @local::def_offset },
                { channel: 7 enable: true dc_offset: @local::def_offset } ]
END_PROLOG
daq: {
  max_fragment_size_words: 2097152
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xDriver
    link_type: PCIE
    link_number: %{link_number}
    io_level: TTL
    acquisition_gate_us: %{gate_width}
    pre_trigger_us: %{pre_trigger}
    acquisition_mode: HW
    multiboard_sincronicazion: true
    #max_buffers_code: 4
    channels: @local::def_channels
    generator_ds50: {
      fragment_id: %{fragment_id}
    }
    rt_priority: 2
  }
}"

V1720_SIM_CONFIG = "\
daq: {
  max_fragment_size_words: 2097152 
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    mpi_sync_interval: 50
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V172xSimulator 
    freqs_file: \"ds50_hist.dat\" 
    fragments_per_event: 1 
    nChannels: 80000 
    generator_ds50: {
      fragment_id: %{fragment_id}
    }
    rt_priority: 2
  }
}"

V1190_CONFIG = "\
daq: {
  max_fragment_size_words: 2097152
  fragment_receiver: {
    mpi_buffer_count: %{buffer_count}
    first_event_builder_rank: %{total_frs}
    event_builder_count: %{total_ebs}
    generator: V1190Driver
    link_type: PCIE
    link_number: %{link_number}
    vme_base: 0x10000000
    generator_ds50: {
      fragment_id: %{fragment_id}
    }
    rt_priority: 2
  }
}"

class ConfigGen
  def generateEventBuilder(ebIndex, totalFRs, totalEBs, totalAGs, compressionLevel, 
                           totalv1720s, totalv1724s, dataDir, onmonEnable,
                           diskWritingEnable)
    # Do the substitutions in the event builder configuration given the options
    # that were passed in from the command line.  
    ebConfig = String.new(EB_CONFIG)
    ebConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    ebConfig.gsub!(/\%\{ag_rank\}/, String(totalFRs + totalEBs))
    ebConfig.gsub!(/\%\{buffer_count\}/, String(totalFRs*8))
    if Integer(compressionLevel) > 0 && Integer(compressionLevel) < 3
      if Integer(totalv1720s) > 0 and Integer(totalv1724s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "")
        ebConfig.gsub!(/\%\{enable_compression\}/, "")
      elsif Integer(totalv1720s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_compression\}/, "")
      elsif Integer(totalv1724s) > 0
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_compression\}/, "")
      else
        ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        ebConfig.gsub!(/\%\{enable_compression\}/, "#")
      end
    else
      ebConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
      ebConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
      ebConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      ebConfig.gsub!(/\%\{enable_compression\}/, "#")
    end
    if Integer(compressionLevel) > 1
      ebConfig.gsub!(/\%\{drop_uncompressed\}/, "")
    else
      ebConfig.gsub!(/\%\{drop_uncompressed\}/, "#")
    end
    if Integer(totalAGs) >= 1
      ebConfig.gsub!(/\%\{netmon_output\}/, "")
      ebConfig.gsub!(/\%\{root_output\}/, "#")

      ebConfig.gsub!(/\%\{output_and_onmon\}/, "#")
      ebConfig.gsub!(/\%\{output_only\}/, "")
      ebConfig.gsub!(/\%\{onmon_only\}/, "#")
    else
      ebConfig.gsub!(/\%\{netmon_output\}/, "#")
      if Integer(diskWritingEnable) != 0 && Integer(onmonEnable) != 0
        ebConfig.gsub!(/\%\{root_output\}/, "")
        ebConfig.gsub!(/\%\{output_and_onmon\}/, "")
        ebConfig.gsub!(/\%\{output_only\}/, "#")
        ebConfig.gsub!(/\%\{onmon_only\}/, "#")
      elsif Integer(onmonEnable) != 0
        ebConfig.gsub!(/\%\{root_output\}/, "#")
        ebConfig.gsub!(/\%\{output_and_onmon\}/, "#")
        ebConfig.gsub!(/\%\{output_only\}/, "#")
        ebConfig.gsub!(/\%\{onmon_only\}/, "")
      elsif Integer(diskWritingEnable) != 0
        ebConfig.gsub!(/\%\{root_output\}/, "")
        ebConfig.gsub!(/\%\{output_and_onmon\}/, "#")
        ebConfig.gsub!(/\%\{output_only\}/, "")
        ebConfig.gsub!(/\%\{onmon_only\}/, "#")
      else
        ebConfig.gsub!(/\%\{root_output\}/, "#")
        ebConfig.gsub!(/\%\{output_and_onmon\}/, "#")
        ebConfig.gsub!(/\%\{output_only\}/, "#")
        ebConfig.gsub!(/\%\{onmon_only\}/, "#")
      end
    end

    currentTime = Time.now
    fileName = "ds50daq_eb%02d_" % ebIndex
    fileName += "%d%02d%02d-%02d%02d%02d" % [currentTime.year, currentTime.month,
                                             currentTime.day, currentTime.hour,
                                             currentTime.min, currentTime.sec]
    fileName += ".root"
    outputFile = File.join(dataDir, fileName)
    ebConfig.gsub!(/\%\{output_file\}/, outputFile)

    return ebConfig
  end

  def generateAggregator(dataDir, runNumber, totalFRs, totalEBs, bunchSize,
                         compressionLevel, totalv1720s, totalv1724s, onmonEnable,
                         diskWritingEnable)
    # Do the substitutions in the aggregator configuration given the options
    # that were passed in from the command line.  Assure that files written out
    # by each AG are unique by including a timestamp in the file name.
    currentTime = Time.now
    fileName = "ds50daq_"
    fileName += "%d%02d%02d-%02d%02d%02d" % [currentTime.year, currentTime.month,
                                             currentTime.day, currentTime.hour,
                                             currentTime.min, currentTime.sec]
    fileName += ".root"
    outputFile = File.join(dataDir, fileName)
    agConfig = String.new(AG_CONFIG)
    agConfig.gsub!(/\%\{output_file\}/, outputFile)
    agConfig.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    agConfig.gsub!(/\%\{total_frs\}/, String(totalFRs))
    agConfig.gsub!(/\%\{buffer_count\}/, String(totalEBs*10))
    agConfig.gsub!(/\%\{bunch_size\}/, String(bunchSize))
    if Integer(compressionLevel) > 0 && Integer(compressionLevel) < 3
      if Integer(totalv1720s) > 0 and Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "")
        agConfig.gsub!(/\%\{enable_compression\}/, "")
      elsif Integer(totalv1720s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_compression\}/, "")
      elsif Integer(totalv1724s) > 0
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_compression\}/, "")
      else
        agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
        agConfig.gsub!(/\%\{enable_compression\}/, "#")
      end
    else
      agConfig.gsub!(/\%\{enable_1720_compression\}/, "#")
      agConfig.gsub!(/\%\{enable_1724_compression\}/, "#")
      agConfig.gsub!(/\%\{enable_172x_compression\}/, "#")
      agConfig.gsub!(/\%\{enable_compression\}/, "#")
    end
    if Integer(compressionLevel) > 1
      agConfig.gsub!(/\%\{drop_uncompressed\}/, "")
    else
      agConfig.gsub!(/\%\{drop_uncompressed\}/, "#")
    end
    if Integer(diskWritingEnable) != 0 && Integer(onmonEnable) != 0
      agConfig.gsub!(/\%\{root_output\}/, "")
      agConfig.gsub!(/\%\{output_and_onmon\}/, "")
      agConfig.gsub!(/\%\{output_only\}/, "#")
      agConfig.gsub!(/\%\{onmon_only\}/, "#")
    elsif Integer(onmonEnable) != 0
      agConfig.gsub!(/\%\{root_output\}/, "#")
      agConfig.gsub!(/\%\{output_and_onmon\}/, "#")
      agConfig.gsub!(/\%\{output_only\}/, "#")
      agConfig.gsub!(/\%\{onmon_only\}/, "")
    elsif Integer(diskWritingEnable) != 0
      agConfig.gsub!(/\%\{root_output\}/, "")
      agConfig.gsub!(/\%\{output_and_onmon\}/, "#")
      agConfig.gsub!(/\%\{output_only\}/, "")
      agConfig.gsub!(/\%\{onmon_only\}/, "#")
    else
      agConfig.gsub!(/\%\{root_output\}/, "#")
      agConfig.gsub!(/\%\{output_and_onmon\}/, "#")
      agConfig.gsub!(/\%\{output_only\}/, "#")
      agConfig.gsub!(/\%\{onmon_only\}/, "#")
    end

    return agConfig  
  end

  def generateV1495(totalEBs, totalFRs, linkNumber, pulserRate, laserRate,
                    randomTriggerEnable, gateWidth, lowThreshold, fnalMode)
    # Do the substitutions in the V1495 configuration given the options that 
    # were passed in from the command line.
    v1495Config = String.new(V1495_CONFIG)
    v1495Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1495Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1495Config.gsub!(/\%\{last_fragment_id\}/, String(totalFRs-1))
    v1495Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1495Config.gsub!(/\%\{pulser_rate\}/, String(pulserRate))
    v1495Config.gsub!(/\%\{laser_rate\}/, String(laserRate))
    if Integer(randomTriggerEnable) > 0
      v1495Config.gsub!(/\%\{random_trigger_enable\}/, "true")
    else
      v1495Config.gsub!(/\%\{random_trigger_enable\}/, "false")
    end
    v1495Config.gsub!(/\%\{low_threshold\}/, String(lowThreshold))
    v1495Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    v1495Config.gsub!(/\%\{gate_width\}/, String(gateWidth+10))
    if Integer(fnalMode) > 0
      v1495Config.gsub!(/\%\{fnal_mode\}/, "")
    else
      v1495Config.gsub!(/\%\{fnal_mode\}/, "#")
    end
    return v1495Config
  end
  
  def generateV1720(v1720Index, totalEBs, totalFRs, linkNumber, sim, gateWidth)
    # Do the substitutions in the V1720 configuration given the options that 
    # were passed in from the command line.
    if sim != 0
      v1720Config = String.new(V1720_SIM_CONFIG)
    else
      v1720Config = String.new(V1720_CONFIG)
    end
    
    v1720Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1720Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1720Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1720Config.gsub!(/\%\{fragment_id\}/, String(v1720Index))
    v1720Config.gsub!(/\%\{gate_width\}/, String(gateWidth))
    v1720Config.gsub!(/\%\{pre_trigger\}/, String(gateWidth/4.0))
    v1720Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    return v1720Config
  end
  
  def generateV1190(v1190Index, totalEBs, totalFRs, linkNumber)
    # Do the substitutions in the V1190 configuration given the options that 
    # were passed in from the command line.
    v1190Config = String.new(V1190_CONFIG)
    v1190Config.gsub!(/\%\{total_ebs\}/, String(totalEBs))
    v1190Config.gsub!(/\%\{total_frs\}/, String(totalFRs))
    v1190Config.gsub!(/\%\{link_number\}/, String(linkNumber))
    v1190Config.gsub!(/\%\{fragment_id\}/, String(totalFRs-2))
    v1190Config.gsub!(/\%\{buffer_count\}/, String(totalEBs*8))
    return v1190Config
  end
end

class CommandLineParser
  def initialize(configGen)
    @configGen = configGen
    @options = OpenStruct.new
    @options.aggregators = []
    @options.eventBuilders = []
    @options.v1720s = []
    @options.v1190s = []
    @options.v1495s = []
    @options.dataDir = nil
    @options.command = nil
    @options.summary = false
    @options.runNumber = "0101"
    @options.serialize = false
    @options.runOnmon = 0
    @options.writeData = 1

    @optParser = OptionParser.new do |opts|
      opts.banner = "Usage: ds50MasterCtrl.rb [options]"
      opts.separator ""
      opts.separator "Specific options:"

      opts.on("--eb [host,port,compression_level]", Array,
              "Add an event builder that runs on the",
              "specified host and port and optionally",
              "compresses ADC data.") do |eb|
        if eb.length != 3
          puts "You must specifiy a host, port, and compression level."
          exit
        end
        ebConfig = OpenStruct.new
        ebConfig.host = eb[0]
        ebConfig.port = Integer(eb[1])
        ebConfig.compression_level = Integer(eb[2])
        ebConfig.kind = "eb"
        ebConfig.index = @options.eventBuilders.length
        @options.eventBuilders << ebConfig
      end

      opts.on("--ag [host,port,bunch_size,compression_level]", Array,
              "Add an aggregator that runs on the",
              "specified host and port.  Also specify the",
              "number of events to pass to art per bunch,",
              "and the compression level.") do |ag|
        if ag.length != 4
          puts "You must specifiy a host, port, bunch size, and"
          puts "compression level."
          exit
        end
        agConfig = OpenStruct.new
        agConfig.host = ag[0]
        agConfig.port = Integer(ag[1])
        agConfig.kind = "ag"
        agConfig.bunch_size = Integer(ag[2])
        agConfig.compression_level = Integer(ag[3])
        @options.aggregators << agConfig
      end
    
      opts.on("--v1495 [host,port,link_number,pulser_rate,laser_rate,random_trigger,gate_width,low_threshold,fnal_mode]", Array, 
              "Add a V1495 fragment receiver that runs on the specified host,",
              "port, and link, and has the specified parameters for pulser rate ",
              "laser rate, random trigger enable, gate width, majority logic ",
              "low threshold, and FNAL mode selection flag.") do |v1495|
        if v1495.length != 9
          puts "You must specifiy a host, port, link_number, pulser rate,"
          puts "laser rate, random trigger enable, gate width,"
          puts "majority logic low threshold, and FNAL mode flag for the V1495."
          exit
        end
        v1495Config = OpenStruct.new
        v1495Config.host = v1495[0]
        v1495Config.port = Integer(v1495[1])
        v1495Config.link_number = Integer(v1495[2])
        v1495Config.pulser_rate = Float(v1495[3])
        v1495Config.laser_rate = Float(v1495[4])
        v1495Config.random_trigger_enable = Integer(v1495[5])
        v1495Config.gate_width = Float(v1495[6])
        v1495Config.low_threshold = Integer(v1495[7])
        v1495Config.fnal_mode = Integer(v1495[8])
        v1495Config.kind = "v1495"
        v1495Config.index = @options.v1495s.length + @options.v1720s.length
        @options.v1495s << v1495Config
      end

      opts.on("--v1720 [host,port,link_number,sim,gate_width]", Array, 
              "Add a V1720 fragment receiver that runs on the",
              "specified host, port, link and gate width.  The sim", 
              "parameter (0 or 1) determines whether or",
              "not to use the V1720 simulator.") do |v1720|
        if v1720.length != 5
          puts "You must specifiy a host, port, link_number, whether or not"
          puts "the V1720 is simulated, and the gate width (usec)."
          exit
        end
        v1720Config = OpenStruct.new
        v1720Config.host = v1720[0]
        v1720Config.port = Integer(v1720[1])
        v1720Config.link_number = Integer(v1720[2])
        v1720Config.sim = Integer(v1720[3])
        v1720Config.gate_width = Integer(v1720[4])
        v1720Config.kind = "v1720"
        v1720Config.index = @options.v1495s.length + @options.v1720s.length
        @options.v1720s << v1720Config
      end

      opts.on("--v1724 [host,port,link_number,sim,gate_width]", Array, 
              "Add a V1724 fragment receiver that runs on the",
              "specified host, port, link and gate width.  The sim", 
              "parameter (0 or 1) determines whether or",
              "not to use the V1724 simulator.") do |v1724|
        if v1724.length != 5
          puts "You must specifiy a host, port, link_number, whether or not"
          puts "the V1724 is simulated, and the gate width [usec]."
          exit
        end
        v1724Config = OpenStruct.new
        v1724Config.host = v1724[0]
        v1724Config.port = Integer(v1724[1])
        v1724Config.link_number = Integer(v1724[2])
        v1724Config.sim = Integer(v1724[3])
        v1724Config.gate_width = Integer(v1724[4])
        v1724Config.kind = "v1724"
        # NOTE that we're simply adding this to the 1720 list...
        @options.v1720s << v1724Config
      end

      opts.on("--v1190 [host,port,link_number]", Array, 
              "Add a V1190 fragment receiver that runs on",
              "the specified host, port and link.") do |v1190|
        if v1190.length != 3
          puts "You must specifiy a host, port, link_number."
          exit
        end
        v1190Config = OpenStruct.new
        v1190Config.host = v1190[0]
        v1190Config.port = Integer(v1190[1])
        v1190Config.link_number = Integer(v1190[2])
        v1190Config.kind = "v1190"
        @options.v1190s << v1190Config
      end

      opts.on("-d", "--data-dir [data dir]", 
              "Directory that the event builders will", "write data to.") do |dataDir|
        @options.dataDir = dataDir
      end

      opts.on("-m", "--online-monitoring [enable flag (0 or 1)]", 
              "Whether to run the online monitoring modules.") do |runOnmon|
        @options.runOnmon = runOnmon
      end

      opts.on("-w", "--write-data [enable flag (0 or 1)]", 
              "Whether to write data to disk.") do |writeData|
        @options.writeData = writeData
      end

      opts.on("-c", "--command [command]", 
              "Execute a command: start, stop, init, shutdown, pause, resume.") do |command|
        @options.command = command
      end

      opts.on("-r", "--run-number [number]", "Specify the run number.") do |run|
        @options.runNumber = run
      end

      opts.on("-s", "--summary", "Summarize the configuration.") do
        @options.summary = true
      end

      opts.on("-e", "--serialize", "Serialize the config for System Control.") do
        @options.serialize = true
      end

      opts.on_tail("-h", "--help", "Show this message.") do
        puts opts
        exit
      end
    end
  end

  def parse()
    if ARGV.length == 0
      puts @optParser
      exit
    end

    @optParser.parse(ARGV)
    self.validate()
  end

  def validate()
    # Make sure link numbers are unique
    # In data mode make sure there is an eb, 1720, 1190 and 1495
    # In sim mode make sure there is an eb and a 1720
    return nil
  end

  def summarize()
    # Print out a summary of the configuration that was passed in from the
    # the command line.  Everything will be printed in terms of what process
    # is running on which host.
    puts "Configuration Summary:"
    hostMap = {}
    (@options.eventBuilders + @options.v1495s + @options.v1720s + 
     @options.aggregators + @options.v1190s).each do |proc|
      if not hostMap.keys.include?(proc.host)
        hostMap[proc.host] = []
      end
      hostMap[proc.host] << proc
    end

    hostMap.each_key { |host|
      puts "  %s:" % [host]
      hostMap[host].sort! { |x,y|
        x.port <=> y.port
      }

      totalEBs = @options.eventBuilders.length
      totalFRs = @options.v1495s.length + @options.v1720s.length
      hostMap[host].each do |item|
        case item.kind
        when "eb"
          puts "    EventBuilder, port %d, rank %d" % [item.port, totalFRs + item.index]
        when "ag"
          puts "    Aggregator, port %d, rank %d" % [item.port, totalEBs + totalFRs]
        when "v1495"
          puts "    FragmentReceiver, V1495, port %s, link %s, rank %d" % [item.port, 
                                                                           item.link_number,
                                                                           item.index]
        when "v1720", "v1724"
          if item.sim != 0
            puts "    FragmentReceiver, Simulated %s, port %d, rank %d" % [item.kind.upcase,
                                                                           item.port,
                                                                           item.index]
          else
            puts "    FragmentReceiver, %s, port %d, link %d, rank %d" % [item.kind.upcase,
                                                                          item.port, 
                                                                          item.link_number,
                                                                          item.index]
          end
        when "v1190"
          puts "    FragmentReceiver, V1190, port %s, link %s" % [item.port, 
                                                                  item.link_number]
        end
      end
      puts ""
    }
    STDOUT.flush
    return nil
  end

  def serialize()
    # The current system control pulls the FHCL config to initialize DAQ
    # processes from a series of files on disk.  The files are named as follows:
    #  initorder#hostname#port#description#on.fcl
    # 
    # This method will generate configurations given the command line options
    # and write out files for System Control.
    puts "Serializing config:"
    procIndex = 1
    v1720Index = 0
    v1190Index = 0
    ebIndex = 0
    totalv1720s = 0
    totalv1724s = 0
    @options.v1720s.each do |proc|
      case proc.kind
      when "v1720"
        totalv1720s += 1
      when "v1724"
        totalv1724s += 1
      end
    end
    totalFRs = @options.v1495s.length + @options.v1720s.length + @options.v1190s.length
    totalEBs = @options.eventBuilders.length
    totalAGs = @options.aggregators.length
    desc = "%dx%d" % [totalFRs, totalEBs]
    (@options.eventBuilders + @options.v1495s + @options.v1720s + \
     @options.v1190s + @options.aggregators).each do |proc|
      fileName = "%d#%s#%d#%s#on.fcl" % [procIndex, proc.host, proc.port, desc]
      procIndex += 1
      puts "  writing %s..." % fileName
      handle = File.open(fileName, "w")
      case proc.kind
      when "eb"
        cfg = @configGen.generateEventBuilder(ebIndex, totalFRs, totalEBs, 
                                              totalAGs, proc.compression_level, 
                                              totalV1720s, totalv1724s,
                                              @options.dataDir, @options.runOnmon,
                                              @options.writeData)
        ebIndex += 1
      when "ag"
        cfg = @configGen.generateAggregator(@options.dataDir, @options.runNumber,
                                            totalFRs, totalEBs, proc.bunch_size,
                                            proc.compression_level,
                                            totalV1720s, totalv1724s,
                                            @options.runOnmon, @options.writeData)
      when "v1495"
        cfg = @configGen.generateV1495(totalEBs, totalFRs, proc.link_number,
                                       proc.pulser_rate, proc.laser_rate,
                                       proc.random_trigger_enable,
                                       proc.gate_width, proc.low_threshold,
                                       proc.fnal_mode)
      when "v1720", "v1724"
        cfg = @configGen.generateV1720(v1720Index, totalEBs, totalFRs, 
                                       proc.link_number, proc.sim, proc.gate_width)
        v1720Index += 1
      when "v1190"
        cfg = @configGen.generateV1190(v1190Index, totalEBs, totalFRs,
                                       proc.link_number)
      end
      handle.write(cfg)
      handle.close()
    end
    STDOUT.flush
  end

  def getOptions()
    return @options
  end
end

class SystemControl
  def initialize(cmdLineOptions, configGen)
    @options = cmdLineOptions
    @configGen = configGen
  end

  def init()
    # The DAQ components need to be initialized in parallel as this is the step
    # in which connections are made between the various components.  The XMLRPC
    # call will block until all connections have been made which means that the
    # configurations need to be posted to all processes at the same time.
    ebIndex = 0
    totalv1720s = 0
    totalv1724s = 0
    @options.v1720s.each do |proc|
      case proc.kind
      when "v1720"
        totalv1720s += 1
      when "v1724"
        totalv1724s += 1
      end
    end
    totalFRs = @options.v1495s.length + @options.v1720s.length + @options.v1190s.length
    totalEBs = @options.eventBuilders.length
    totalAGs = @options.aggregators.length

    # 27-Jun-2013, KAB - send INIT to 1495 first
    threads = []
    @options.v1495s.each { |v1495Options|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, v1495Options.host, v1495Options.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(v1495Options.host, "/RPC2", 
                                          v1495Options.port)
        cfg = @configGen.generateV1495(totalEBs, totalFRs, 
                                       v1495Options.link_number,
                                       v1495Options.pulser_rate,
                                       v1495Options.laser_rate,
                                       v1495Options.random_trigger_enable,
                                       v1495Options.gate_width,
                                       v1495Options.low_threshold,
                                       v1495Options.fnal_mode)
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: V1495 FragmentReceiver on %s:%d result: %s" %
          [currentTime, v1495Options.host, v1495Options.port, result]
        STDOUT.flush
      end
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }

    # 27-Jun-2013, KAB - send INIT to other hardware next
    threads = []
    v1720Index = 0
    @options.v1720s.each { |v1720Options|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, v1720Options.host, v1720Options.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(v1720Options.host, "/RPC2",
                                          v1720Options.port)
        cfg = @configGen.generateV1720(v1720Index, totalEBs, totalFRs, 
                                       v1720Options.link_number,
                                       v1720Options.sim, v1720Options.gate_width)
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: V1720 FragmentReceiver on %s:%d result: %s" %
          [currentTime, v1720Options.host, v1720Options.port, result]
        STDOUT.flush
      end
      v1720Index += 1
    }
    v1190Index = 0
    @options.v1190s.each { |v1190Options|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, v1190Options.host, v1190Options.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(v1190Options.host, "/RPC2",
                                          v1190Options.port)
        cfg = @configGen.generateV1190(v1190Index, totalEBs, totalFRs, 
                                       v1190Options.link_number)
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: V1190 FragmentReceiver on %s:%d result: %s" %
          [v1190Options.host, v1190Options.port, result]
        STDOUT.flush
      end
      v1190Index += 1      
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }

    # 27-Jun-2013, KAB - send INIT to EBs and AG last
    threads = []
    @options.eventBuilders.each { |ebOptions|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, ebOptions.host, ebOptions.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(ebOptions.host, "/RPC2", 
                                          ebOptions.port)
        cfg = @configGen.generateEventBuilder(ebIndex, totalFRs, totalEBs, totalAGs,
                                              ebOptions.compression_level,
                                              totalv1720s, totalv1724s,
                                              @options.dataDir, @options.runOnmon,
                                              @options.writeData)
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: EventBuilder on %s:%d result: %s" %
          [currentTime, ebOptions.host, ebOptions.port, result]
        STDOUT.flush
      end
      ebIndex += 1
    }
    @options.aggregators.each { |agOptions|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Sending the INIT command to %s:%d." %
        [currentTime, agOptions.host, agOptions.port]
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(agOptions.host, "/RPC2", 
                                          agOptions.port)
        cfg = @configGen.generateAggregator(@options.dataDir, @options.runNumber,
                                            totalFRs, totalEBs, agOptions.bunch_size,
                                            agOptions.compression_level,
                                            totalv1720s, totalv1724s,
                                            @options.runOnmon, @options.writeData)
        result = xmlrpcClient.call("daq.init", cfg)
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        puts "%s: Aggregator on %s:%d result: %s" %
          [currentTime, agOptions.host, agOptions.port, result]
        STDOUT.flush
      end
    }
    STDOUT.flush
    threads.each { |aThread|
      aThread.join()
    }
  end

  def start(runNumber)
    # Starting runs is simple in that all processes are sent the same message
    # and that can be done serially.  The V1495 needs to be started last as we
    # don't want other components missing triggers if they are initialized after
    # the V1495.
    (@options.aggregators + @options.eventBuilders + @options.v1720s + 
     @options.v1190s + @options.v1495s).each do |proc|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Attempting to connect to %s:%d and start a run." %
        [currentTime, proc.host, proc.port]
      STDOUT.flush
      xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
      result = xmlrpcClient.call("daq.start", runNumber)
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      case proc.kind
      when "eb"
        puts "%s: EventBuilder on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "ag"
        puts "%s: Aggregator on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "v1495"
        puts "%s: V1495 FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "v1720", "v1724"
        puts "%s: %s FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.kind.upcase, proc.host, proc.port, result]
      when "v1190"
        puts "%s: V1190 FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      end
      STDOUT.flush
    end
  end

  def sendSerialCommand(commandName, procs)
    procs.each do |proc|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Attempting to connect to %s:%d and %s a run." %
        [currentTime, proc.host, proc.port, commandName]
      STDOUT.flush
      xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
      xmlrpcClient.timeout = 60
      if commandName == "stop"
        if proc.kind == "ag"
          xmlrpcClient.timeout = 120
        elsif proc.kind == "eb"
          xmlrpcClient.timeout = 45
        else
          xmlrpcClient.timeout = 30
        end
      end
      begin
        result = xmlrpcClient.call("daq.%s" % [commandName])
      rescue Exception => msg
        result = "Exception: " + msg
      end
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      case proc.kind
      when "eb"
        puts "%s: EventBuilder on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "ag"
        puts "%s: Aggregator on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "v1495"
        puts "%s: V1495 FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "v1720"
        puts "%s: V1720 FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      when "v1190"
        puts "%s: V1190 FragmentReceiver on %s:%d result: %s" %
          [currentTime, proc.host, proc.port, result]
      end
      STDOUT.flush
    end
  end

  def sendCommandSet(commandName, procs)
    threads = []
    procs.each do |proc|
      currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
      puts "%s: Attempting to connect to %s:%d and %s a run." %
        [currentTime, proc.host, proc.port, commandName]
      STDOUT.flush
      threads << Thread.new() do
        xmlrpcClient = XMLRPC::Client.new(proc.host, "/RPC2", proc.port)
        xmlrpcClient.timeout = 60
        if commandName == "stop"
          if proc.kind == "ag"
            xmlrpcClient.timeout = 120
          elsif proc.kind == "eb"
            xmlrpcClient.timeout = 45
          else
            xmlrpcClient.timeout = 30
          end
        end
        begin
          result = xmlrpcClient.call("daq.%s" % [commandName])
        rescue Exception => msg
          result = "Exception: " + msg
        end
        currentTime = DateTime.now.strftime("%Y/%m/%d %H:%M:%S")
        case proc.kind
        when "eb"
          puts "%s: EventBuilder on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "ag"
          puts "%s: Aggregator on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "v1495"
          puts "%s: V1495 FragmentReceiver on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "v1720"
          puts "%s: V1720 FragmentReceiver on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        when "v1190"
          puts "%s: V1190 FragmentReceiver on %s:%d result: %s" %
            [currentTime, proc.host, proc.port, result]
        end
        STDOUT.flush
      end
    end
    threads.each { |aThread|
      aThread.join()
    }
  end

  def shutdown()
    self.sendSerialCommand("shutdown", @options.aggregators + \
                           @options.eventBuilders + @options.v1495s + 
                           @options.v1720s + @options.v1190s)
  end

  def pause()
    self.sendSerialCommand("pause", @options.v1495s + @options.v1720s + \
                           @options.v1190s + @options.eventBuilders + \
                           @options.aggregators)
  end

  def stop()
    self.sendCommandSet("stop", @options.v1495s)
    self.sendCommandSet("stop", @options.v1720s + @options.v1190s)
    self.sendCommandSet("stop", @options.eventBuilders)
    self.sendCommandSet("stop", @options.aggregators)
  end

  def resume()
    self.sendSerialCommand("resume", @options.aggregators + \
                           @options.eventBuilders + @options.v1495s + 
                           @options.v1720s + @options.v1190s)
  end
end

if __FILE__ == $0
  cfgGen = ConfigGen.new
  cmdLineParser = CommandLineParser.new(cfgGen)
  cmdLineParser.parse()
  options = cmdLineParser.getOptions()

  if options.summary
    cmdLineParser.summarize()
  end

  if options.serialize
    cmdLineParser.serialize()
    exit
  end

  sysCtrl = SystemControl.new(options, cfgGen)

  if options.command == "init"
    sysCtrl.init()
  elsif options.command == "start"
    sysCtrl.start(options.runNumber)
  elsif options.command == "stop"
    sysCtrl.stop()
  elsif options.command == "shutdown"
    sysCtrl.shutdown()
  elsif options.command == "pause"
    sysCtrl.pause()
  elsif options.command == "resume"
    sysCtrl.resume()
  end
end
